import asyncio
import copy
import datetime
import logging
import socket
import threading
import time
import traceback
from collections import defaultdict
from enum import Enum
from urllib.parse import parse_qs
from wsgiref.simple_server import make_server

from prometheus_client.exposition import (
    ThreadingWSGIServer,
    _bake_output,
    _SilentHandler,
)
from prometheus_client.metrics import Counter, Gauge, Histogram
from prometheus_client.registry import REGISTRY

from .metrics import (
    BUCKETS_HIST,
    OPERATION_EXPAND_METRICS,
    PROMETHEUS_METRICS_REGISTRY,
)


class TaskStatus(Enum):
    OK = 1
    FATAL = 2
    CONNECTION_ERROR = 3
    OTHER_EXCEPTION = -1
    INTERNAL = 4
    UNKNOWN = 5


class ErrorParser:
    _types = []
    _taskstatus2types = {}

    @staticmethod
    def init():
        for status_key, entry in MonitorServer.errors.items():
            ErrorParser._taskstatus2types[status_key] = entry["service_errors"]
            ErrorParser._types.extend(entry["service_errors"])

    @staticmethod
    def parse(error):
        traceback_string = traceback.format_exc().lower()
        for error_type_name in ErrorParser._types:
            if error_type_name == "other":
                continue
            if error_type_name in traceback_string:
                return error_type_name
        return "other"

    @staticmethod
    async def set_status(error_type=None):
        server = MonitorServer.get_instance()
        if not server:
            return
        if error_type is None:
            await server.set_task_status(TaskStatus.OK)
            return
        mapped_task_status = None
        for status_key, err_list in ErrorParser._taskstatus2types.items():
            if error_type in err_list:
                mapped_task_status = (
                    getattr(TaskStatus, status_key,
                            None) or TaskStatus.OTHER_EXCEPTION
                )
                break
        if mapped_task_status is None:
            mapped_task_status = TaskStatus.OK
        await server.set_task_status(mapped_task_status)


class StatusParser:
    _types = ["failure", "success"]

    @staticmethod
    def parse(error=False):
        return StatusParser._types[0] if error else StatusParser._types[1]


class MonitorStorage:
    """Async storage for monitoring state. Registry is updated only when /metrics is requested."""

    @staticmethod
    def _sanitize_metric_name(name):
        return name.replace(" ", "_").replace("-", "_").replace(".", "_").lower()

    def _gauge_get(self, metric):
        """Return current value of a gauge/counter with default labels (for combining metrics)."""
        return metric.labels(*MonitorServer._get_labels())._value.get()

    def __init__(self):
        self._lock = threading.Lock()
        self._stats = {}
        self._status_series = {}
        self._init_time = str(datetime.datetime.now())

    def get_time(self):
        return self._init_time

    def refresh_registry_for_metrics(self):
        """Update PROMETHEUS_METRICS_REGISTRY from current stats. Call from MetricServer on /metrics request (e.g. every 30s)."""
        with self._lock:
            snapshot = copy.deepcopy(self._stats)
        self._refresh_registry_from_stats(snapshot)

    @staticmethod
    def _get_aggregate_stats(all_stats, show_tasks=True, last_count=-1):
        aggregated_by_key = (
            {"tasks": [len(all_stats.get("tasks", {}))]} if show_tasks else {}
        )
        stats_series = all_stats.get("stats", [])
        if last_count > 0:
            stats_series = stats_series[-last_count:]
        for stats_point in stats_series:
            for stat_key in stats_point:
                aggregated_by_key.setdefault(stat_key, []).append(stats_point[stat_key])
        for aggregation_key in aggregated_by_key:
            try:
                aggregated_by_key[aggregation_key] = (
                    sum(aggregated_by_key[aggregation_key]) * 1.0
                    / len(aggregated_by_key[aggregation_key])
                )
            except Exception:
                aggregated_by_key[aggregation_key] = 0.0
        return aggregated_by_key

    def _apply_operations(
        self,
        values_to_aggregate,
        operation_names=None,
        prefix="",
        postfix="",
        skip_op_in_suffix=None,
    ):
        operation_names = operation_names or ["sum"]
        for operation_name in operation_names:
            aggregation_function = OPERATION_EXPAND_METRICS.get(operation_name)
            if aggregation_function is None:
                raise NotImplementedError(
                    "Operation %s not implemented" % operation_name
                )
            operation_name_suffix_for_metric = (
                "_" if operation_name == skip_op_in_suffix else "_%s_" % operation_name
            )
            prometheus_metric_key = prefix + operation_name_suffix_for_metric + postfix
            if prometheus_metric_key not in PROMETHEUS_METRICS_REGISTRY:
                PROMETHEUS_METRICS_REGISTRY[prometheus_metric_key] = Gauge(
                    prometheus_metric_key,
                    prometheus_metric_key.replace("_", " "),
                    MonitorServer._label_names(),
                )
            PROMETHEUS_METRICS_REGISTRY[prometheus_metric_key].labels(
                *MonitorServer._get_labels()
            ).set(aggregation_function(values_to_aggregate))

    def _prom_set_value(self, stat_key, numeric_value, prefix):
        key_parts = stat_key.split("___")
        metric_name_from_stat_key = key_parts[0]
        prometheus_label_names = []
        prometheus_label_values = []
        for label_part in key_parts[1:]:
            if len(label_part.split("__")) != 2:
                logging.warning(
                    "Label %s has bad format. Must be <label2name__label2value>.",
                    label_part,
                )
            else:
                label_name_part, label_value_part = label_part.split("__", 1)
                prometheus_label_names.append(label_name_part)
                prometheus_label_values.append(label_value_part)
        prometheus_metric_key_sanitized = prefix + self._sanitize_metric_name(
            metric_name_from_stat_key
        )
        if prometheus_metric_key_sanitized not in PROMETHEUS_METRICS_REGISTRY:
            PROMETHEUS_METRICS_REGISTRY[prometheus_metric_key_sanitized] = Gauge(
                prometheus_metric_key_sanitized,
                "This is metric `%s` from MonitorServer." % metric_name_from_stat_key,
                MonitorServer._label_names(prometheus_label_names),
            )
        PROMETHEUS_METRICS_REGISTRY[prometheus_metric_key_sanitized].labels(
            *MonitorServer._get_labels(prometheus_label_values)
        ).set(numeric_value)

    def _refresh_registry_from_stats(self, all_stats):
        for aggregation_key, aggregation_value in list(
            self._get_aggregate_stats(all_stats, False).items()
        ):
            if isinstance(aggregation_value, (float, int)):
                self._prom_set_value(aggregation_key, aggregation_value, "stats_aggregation_")
        for aggregation_key, aggregation_value in list(
            self._get_aggregate_stats(all_stats, False, 5).items()
        ):
            if isinstance(aggregation_value, (float, int)):
                self._prom_set_value(
                    aggregation_key, aggregation_value, "stats_aggregation_last_"
                )
        for top_level_stat_key in all_stats:
            if top_level_stat_key in ("tasks", "stats"):
                continue
            if isinstance(all_stats[top_level_stat_key], (int, float)):
                self._prom_set_value(
                    top_level_stat_key, all_stats[top_level_stat_key], "stats_"
                )
        task_id_to_start_timestamp = all_stats.get("tasks", [])
        self._apply_operations(
            task_id_to_start_timestamp,
            ["len"],
            prefix="tasks",
            postfix="count",
            skip_op_in_suffix="len",
        )
        current_timestamp_seconds = time.time()
        self._apply_operations(
            [
                (current_timestamp_seconds - task_id_to_start_timestamp[task_id])
                for task_id in task_id_to_start_timestamp
            ],
            ["sum", "max"],
            prefix="current_tasks",
            postfix="time_spend",
            skip_op_in_suffix="sum",
        )
        combined_time_spend_metric_key = "stats_tasks_with_current_time_spend_total"
        current_val = self._gauge_get(PROMETHEUS_METRICS_REGISTRY["current_tasks_time_spend"])
        spent_val = self._gauge_get(PROMETHEUS_METRICS_REGISTRY["stats_tasks_time_spend"])
        PROMETHEUS_METRICS_REGISTRY[combined_time_spend_metric_key].labels(
            *MonitorServer._get_labels()
        ).set(current_val + spent_val)

    def _record_histogram_observation(
        self,
        histogram_metric_key_string,
        observed_value,
        metric_prefix,
        bucket_type="time",
    ):
        key_parts = histogram_metric_key_string.split("___")
        metric_name_part, label_parts = key_parts[0], key_parts[1:]
        histogram_label_names = []
        histogram_label_values = []
        for label_part in label_parts:
            if len(label_part.split("__")) != 2:
                logging.warning(
                    "The label %s is not in the correct format. It should be <label2name__label2value>.",
                    label_part,
                )
            else:
                label_name_part, label_value_part = label_part.split("__", 1)
                histogram_label_names.append(label_name_part)
                histogram_label_values.append(label_value_part)
        sanitized_histogram_metric_key = metric_prefix + self._sanitize_metric_name(
            metric_name_part
        )
        if sanitized_histogram_metric_key not in PROMETHEUS_METRICS_REGISTRY:
            PROMETHEUS_METRICS_REGISTRY[sanitized_histogram_metric_key] = Histogram(
                sanitized_histogram_metric_key,
                "this is hist metric `%s` in stats from MonitorServers"
                % metric_name_part,
                MonitorServer._label_names(histogram_label_names),
                buckets=BUCKETS_HIST[bucket_type],
            )
        PROMETHEUS_METRICS_REGISTRY[sanitized_histogram_metric_key].labels(
            *MonitorServer._get_labels(histogram_label_values)
        ).observe(observed_value)

    async def inc_stats(self, stat_key, count=1):
        with self._lock:
            self._stats[stat_key] = self._stats.get(stat_key, 0) + count
            result = self._stats[stat_key]
        return result

    async def set_stats(self, stat_key, value):
        with self._lock:
            self._stats[stat_key] = value
        return value

    async def set_task_status(self, status, task_type="default"):
        with self._lock:
            old_status, old_series = self._status_series.get(
                task_type, (TaskStatus.OK.name, 0)
            )
            if old_status == status.name:
                new_status, new_series = old_status, old_series + 1
            else:
                new_status, new_series = status.name, 1
            self._status_series[task_type] = (new_status, new_series)
            self._stats["series." + task_type] = (status.name, new_series)

    async def append_stats(self, stat_key, stat_value, max_count=-1):
        with self._lock:
            if stat_key not in self._stats:
                self._stats[stat_key] = []
            self._stats[stat_key].append(stat_value)
            if max_count >= 0:
                self._stats[stat_key] = self._stats[stat_key][-max_count:]
            if isinstance(stat_value, dict):
                for histogram_stat_key, histogram_observed_value in stat_value.items():
                    bucket_type = "time" if "time" in histogram_stat_key else "size"
                    if "time" in histogram_stat_key or "size" in histogram_stat_key:
                        self._record_histogram_observation(
                            histogram_stat_key,
                            histogram_observed_value,
                            "stats_hist_",
                            bucket_type=bucket_type,
                        )
            elif ("time" in stat_key or "size" in stat_key) and isinstance(
                stat_value, (int, float)
            ):
                self._record_histogram_observation(
                    stat_key,
                    stat_value,
                    "stats_hist_",
                    bucket_type="time" if "time" in stat_key else "size",
                )

    async def add_key_value(self, section_key, key_value_tuple):
        with self._lock:
            if section_key not in self._stats:
                self._stats[section_key] = {}
            if section_key == "tasks":
                PROMETHEUS_METRICS_REGISTRY["started_total"].labels(
                    *MonitorServer._get_labels()
                ).inc()
            self._stats[section_key][key_value_tuple[0]] = key_value_tuple[1]

    async def del_key_value(self, section_key, key_to_remove):
        with self._lock:
            if section_key in self._stats and key_to_remove in self._stats[section_key]:
                if section_key == "tasks":
                    PROMETHEUS_METRICS_REGISTRY["stats_tasks_time_spend"].labels(
                        *MonitorServer._get_labels()
                    ).inc(time.time() - self._stats["tasks"][key_to_remove])
                    PROMETHEUS_METRICS_REGISTRY["completed_total"].labels(
                        *MonitorServer._get_labels()
                    ).inc()
                del self._stats[section_key][key_to_remove]

    async def get_stats(self):
        with self._lock:
            return copy.deepcopy(self._stats)

    async def get_statuses(self):
        with self._lock:
            return copy.deepcopy(self._status_series)


class MetricServer:
    """HTTP server in a thread: serves /metrics. Updates registry from storage only when /metrics is requested."""

    def __init__(self, registry=REGISTRY, disable_compression=False, storage=None):
        self._registry = registry
        self._disable_compression = disable_compression
        self._storage = storage

    def _bake_response(self, environ):
        """Build (status, headers, output) for /metrics from self._registry."""
        return _bake_output(
            self._registry,
            environ.get("HTTP_ACCEPT"),
            environ.get("HTTP_ACCEPT_ENCODING"),
            parse_qs(environ.get("QUERY_STRING", "")),
            self._disable_compression,
        )

    def get_wsgi_app(self):
        """Return WSGI app: on /metrics refreshes registry from storage then bakes output."""
        storage = self._storage

        def wsgi_application(environ, start_response):
            if environ["PATH_INFO"] != "/metrics":
                start_response("503 Service Unavailable", [("", "")])
                return [b""]
            if storage:
                storage.refresh_registry_for_metrics()
            status, headers, output = self._bake_response(environ)
            start_response(status, headers)
            return [output]

        return wsgi_application

    def run_in_thread(self, port):
        """Start HTTP server in a daemon thread. Returns the server object."""
        wsgi_app = self.get_wsgi_app()
        while True:
            try:
                http_server = make_server(
                    "",
                    port,
                    wsgi_app,
                    ThreadingWSGIServer,
                    handler_class=_SilentHandler,
                )
                break
            except socket.error:
                time.sleep(1)
        logging.info("Start metric server on port: %s.", port)
        thread = threading.Thread(target=http_server.serve_forever)
        thread.daemon = True
        thread.start()
        return http_server


class MonitorServer:
    """Single entry point: forwards all requests to MonitorStorage. Runs MetricServer in a thread."""

    __instance = None
    __time = None

    @staticmethod
    def _get_labels(extra=None):
        return [MonitorServer.get_time()] + (extra or [])

    @staticmethod
    def _label_names(extra=None):
        return ["inittime"] + (extra or [])

    @staticmethod
    def get_instance():
        return MonitorServer.__instance

    @staticmethod
    def get_time():
        return MonitorServer.__time

    async def wait_for_ending_tasks(self):
        """Wait until no tasks are in progress. Use: await server.wait_for_ending_tasks()."""
        while True:
            stats = await self.get_stats()
            if not stats.get("tasks"):
                return 0
            await asyncio.sleep(1)

    def __init__(self):
        if MonitorServer.__instance:
            raise Exception("This class is a singleton!")
        self._storage = MonitorStorage()
        MonitorServer.__instance = self
        MonitorServer.__time = self._storage.get_time()

    def set_stat_sync(self, stat_key, value):
        """Set one stat key (sync, for init from sync context)."""
        if self._storage:
            with self._storage._lock:
                self._storage._stats[stat_key] = value

    def _init_metrics(self):
        PROMETHEUS_METRICS_REGISTRY["started_total"] = Counter(
            "started", "Total number of started tasks.", MonitorServer._label_names()
        )
        PROMETHEUS_METRICS_REGISTRY["started_total"].labels(
            *MonitorServer._get_labels()).inc(0)
        PROMETHEUS_METRICS_REGISTRY["completed_total"] = Counter(
            "completed", "Total number of completed tasks.", MonitorServer._label_names()
        )
        PROMETHEUS_METRICS_REGISTRY["completed_total"].labels(
            *MonitorServer._get_labels()).inc(0)
        PROMETHEUS_METRICS_REGISTRY["tasks_count"] = Gauge(
            "tasks_count", "The total number of tasks currently being processed.", MonitorServer._label_names()
        )
        PROMETHEUS_METRICS_REGISTRY["tasks_count"].labels(*MonitorServer._get_labels()).inc(0)
        PROMETHEUS_METRICS_REGISTRY["current_tasks_time_spend"] = Gauge(
            "current_tasks_time_spend",
            "Total time spent processing current tasks.",
            MonitorServer._label_names(),
        )
        PROMETHEUS_METRICS_REGISTRY["current_tasks_max_time_spend"] = Gauge(
            "current_tasks_max_time_spend",
            "Maximum processing time for current tasks.",
            MonitorServer._label_names(),
        )
        PROMETHEUS_METRICS_REGISTRY["current_tasks_time_spend"].labels(
            *MonitorServer._get_labels()
        ).inc(0)
        PROMETHEUS_METRICS_REGISTRY["stats_tasks_time_spend"] = Counter(
            "stats_tasks_time_spend",
            "Total time spent on completed tasks.",
            MonitorServer._label_names(),
        )
        PROMETHEUS_METRICS_REGISTRY["stats_tasks_time_spend"].labels(
            *MonitorServer._get_labels()
        ).inc(0)
        PROMETHEUS_METRICS_REGISTRY[
            "stats_tasks_with_current_time_spend_total"
        ] = Gauge(
            "stats_tasks_with_current_time_spend_total",
            "Total time spent processing current and previous tasks.",
            MonitorServer._label_names(),
        )
        PROMETHEUS_METRICS_REGISTRY["stats_tasks_with_current_time_spend_total"].labels(
            *MonitorServer._get_labels()
        ).inc(0)
        PROMETHEUS_METRICS_REGISTRY[
            "stats_tasks_time_spent_quantile"
        ] = Histogram(
            "tasks_time_spent_quantile",
            "Time spent on completing tasks.",
            MonitorServer._label_names(),
            buckets=BUCKETS_HIST["time"],
        )
        for service_info_key in MonitorServer.service_info:
            PROMETHEUS_METRICS_REGISTRY[service_info_key] = Gauge(
                service_info_key,
                MonitorServer.service_info[service_info_key]["info"],
                MonitorServer._label_names(),
            )
            PROMETHEUS_METRICS_REGISTRY[service_info_key].labels(*MonitorServer._get_labels()).set(
                float(MonitorServer.service_info[service_info_key]["value"])
            )

    def run(self, server_config=None, max_workers=1, hpa_max_workers=None):
        server_config = server_config or {}
        http_port = int(server_config.get("port", 9999))
        configured_max_workers = (
            hpa_max_workers if hpa_max_workers is not None else max_workers
        )
        MonitorServer.async_service_threads = defaultdict(
            lambda: asyncio.Semaphore(max_workers)
        )
        MonitorServer.service_threads = defaultdict(
            lambda: threading.Semaphore(max_workers)
        )
        MonitorServer.service_info = dict(
            server_config.get("service info", {}))
        MonitorServer.service_info["service_threads"] = {
            "info": "number of service threads setups by the config",
            "value": configured_max_workers,
        }
        MonitorServer.series_unhealthy = {}
        MonitorServer.errors = {
            "CONNECTION_ERROR": {"service_errors": [], "unhealthy": -1},
            "FATAL": {"service_errors": [], "unhealthy": -1},
            "INTERNAL": {"service_errors": [], "unhealthy": -1},
            "OTHER_EXCEPTION": {"service_errors": ["other"], "unhealthy": -1},
            "UNKNOWN": {"service_errors": [], "unhealthy": -1},
        }
        MonitorServer.errors.update(server_config.get("errors", {}))
        for task_status_enum in TaskStatus:
            if task_status_enum == TaskStatus.OK:
                continue
            MonitorServer.series_unhealthy[task_status_enum.name] = int(
                MonitorServer.errors[task_status_enum.name]["unhealthy"]
            )
        MonitorServer.__time = self._storage.get_time()
        self._init_metrics()
        MetricServer(registry=REGISTRY, storage=self._storage).run_in_thread(http_port)

    async def inc_stats(self, stat_key, value=1):
        if self._storage:
            await self._storage.inc_stats(stat_key, value)
        else:
            logging.warning("The monitor server is not initialized.")

    async def set_task_status(self, status, task_type="default"):
        if self._storage:
            await self._storage.set_task_status(status, task_type=task_type)
        else:
            logging.warning("The monitor server is not initialized.")

    async def append_stats(self, stat_key, stat_value, max_count=-1):
        if self._storage:
            await self._storage.append_stats(stat_key, stat_value, max_count=max_count)
        else:
            logging.warning("The monitor server is not initialized.")

    async def add_key_value(self, section_key, key_value_tuple):
        if self._storage:
            await self._storage.add_key_value(section_key, key_value_tuple)
        else:
            logging.warning("The monitor server is not initialized.")

    async def del_key_value(self, section_key, key_to_remove):
        if self._storage:
            await self._storage.del_key_value(section_key, key_to_remove)
        else:
            logging.warning("The monitor server is not initialized.")

    async def get_stats(self):
        if self._storage:
            return await self._storage.get_stats()
        logging.warning("The monitor server is not initialized.")
        return {}

    async def get_statuses(self):
        if self._storage:
            return await self._storage.get_statuses()
        logging.warning("The monitor server is not initialized.")
        return {}


import logging
import time
import traceback
import uuid
from functools import wraps
from pickle import dumps

from ..metrics import PROMETHEUS_METRICS_REGISTRY
from ..monitoring import ErrorParser, MonitorServer, StatusParser


def grpc_init_wrapper(func):
    @wraps(func)
    def init_and_run(*args, **kwargs):
        ErrorParser.init()
        server = MonitorServer.get_instance()
        service_instance = args[0]
        if server:
            for name in dir(service_instance):
                if name.startswith("_") or not callable(getattr(service_instance, name)):
                    continue
                server.set_stat_sync("request___method__%s" % name, 0)
                for outcome_status in StatusParser._types:
                    server.set_stat_sync(
                        "processed___method__%s___status__%s" % (name, outcome_status), 0
                    )
                for error_kind in ErrorParser._types:
                    server.set_stat_sync(
                        "errors___method__%s___type__%s" % (name, error_kind), 0
                    )
                server.set_stat_sync("request_size___method__%s" % name, 0)
                server.set_stat_sync("response_size___method__%s" % name, 0)
        func(*args, **kwargs)

    return init_and_run


def aio_grpc_method_wrapper(func):
    @wraps(func)
    async def run_with_metrics(*args, **kwargs):
        rpc_method_name = getattr(func, "__name__", "UnknownObject")
        server = MonitorServer.get_instance()
        request_trace_id = str(uuid.uuid4())
        request_start_timestamp_seconds = time.time()

        if server:
            await server.inc_stats("request___method__%s" % rpc_method_name)
            await server.add_key_value("tasks", (request_trace_id, request_start_timestamp_seconds))

        kwargs["info"] = {"context": args[2]}
        request_counters = {"request_size___method__%s" % rpc_method_name: len(dumps(args[1]))}
        grpc_servicer_context = args[2]
        if grpc_servicer_context is not None and "FakeContext" not in str(type(grpc_servicer_context)) and getattr(grpc_servicer_context, "time_remaining", lambda: None)() is not None:
            deadline_remaining_seconds = grpc_servicer_context.time_remaining()
            request_counters["time_remaining_time___method__%s" % rpc_method_name] = deadline_remaining_seconds
            kwargs["info"]["time_remaining"] = lambda: max(0, deadline_remaining_seconds - (time.time() - request_start_timestamp_seconds)) if time.time() - request_start_timestamp_seconds < deadline_remaining_seconds else 0
        else:
            kwargs["info"]["time_remaining"] = lambda: None

        custom_metrics_from_handler = {}
        kwargs["metrics"] = custom_metrics_from_handler
        request_failed = False
        error_status_for_health = None
        try:
            with PROMETHEUS_METRICS_REGISTRY["stats_tasks_time_spent_quantile"].labels(*MonitorServer._get_labels()).time():
                async with MonitorServer.async_service_threads[rpc_method_name]:
                    handler_result = await func(*args, **kwargs)
            request_counters["response_size___method__%s" % rpc_method_name] = len(dumps(handler_result))
            return handler_result
        except Exception as caught_exception:
            logging.error(traceback.format_exc())
            request_failed = True
            error_status_for_health = ErrorParser.parse(caught_exception)
            if server:
                await server.inc_stats("errors___method__%s___type__%s" % (rpc_method_name, error_status_for_health))
            raise
        finally:
            if server:
                await server.del_key_value("tasks", request_trace_id)
            for metric_name, metric_value in custom_metrics_from_handler.items():
                request_counters["%s___method__%s" % (metric_name, rpc_method_name)] = metric_value
            request_counters["time full___method__%s" % rpc_method_name] = time.time() - request_start_timestamp_seconds
            if grpc_servicer_context is not None and "FakeContext" not in str(type(grpc_servicer_context)) and getattr(grpc_servicer_context, "time_remaining", lambda: None)() is not None:
                request_counters["time_remaining_after_time___method__%s" % rpc_method_name] = grpc_servicer_context.time_remaining()
            if server:
                await server.inc_stats("processed___method__%s___status__%s" % (rpc_method_name, StatusParser.parse(request_failed)))
                await server.append_stats("stats", request_counters, 100)
            await ErrorParser.set_status(error_status_for_health)

    return run_with_metrics



import threading
from concurrent import futures

from grpc_health.v1 import health as health_servicer_module
from grpc_health.v1 import health_pb2, health_pb2_grpc

HEALTH_DESCRIPTOR = health_pb2.DESCRIPTOR


def _toggle_health(health_servicer, service):
    health_servicer.set(service, health_pb2.HealthCheckResponse.SERVING)


def _configure_health_server(server, servicer):
    health_servicer = health_servicer_module.HealthServicer(
        experimental_non_blocking=True,
        experimental_thread_pool=futures.ThreadPoolExecutor(max_workers=10),
    )
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
    threading.Thread(
        target=_toggle_health, args=(health_servicer, servicer), daemon=True
    ).start()



import logging

from pythonjsonlogger import jsonlogger


def init_console_log(level: str = "INFO") -> logging.Logger:
    for logger_name in logging.root.manager.loggerDict:
        logging.getLogger(logger_name).setLevel(logging.CRITICAL)

    logger = logging.getLogger()
    logger.handlers.clear()
    logger.setLevel(0)

    console_handler = logging.StreamHandler()
    numeric_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Incorrect logging level.")
    console_handler.setLevel(numeric_level)

    format_string = " ".join(
        "%(" + log_field_name + ")s"
        for log_field_name in [
            "asctime",
            "created",
            "filename",
            "funcName",
            "levelname",
            "lineno",
            "module",
            "message",
            "process",
            "thread",
            "threadName",
        ]
    )
    console_handler.setFormatter(jsonlogger.JsonFormatter(format_string))
    logger.addHandler(console_handler)
    logger.info("A console logger has been created.")
    return logger



import sys

if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version

try:
    __version__ = version("redup-servicekit")
except Exception:
    __version__ = "0.0.0"


__all__ = ['__version__']


from typing import Tuple
from urllib.parse import urlparse

import grpc

DEFAULT_MAX_MESSAGE_LENGTH = 100 * 1024 * 1024

COMPRESSION = {
    "gzip": grpc.Compression.Gzip,
    "deflate": grpc.Compression.Deflate,
    "no_compression": grpc.Compression.NoCompression,
}

KEEPALIVE_MS = 10000
KEEPALIVE_TIMEOUT_MS = 5000
KEEPALIVE_PERMIT_WITHOUT_CALLS = 1
HTTP2_MAX_PINGS_WITHOUT_DATA = 0


class BasicAsyncClient:
    @staticmethod
    def _channel_options(max_message_length: int):
        return [
            ("grpc.max_send_message_length", max_message_length),
            ("grpc.max_message_length", max_message_length),
            ("grpc.max_receive_message_length", max_message_length),
            ("grpc.keepalive_time_ms", KEEPALIVE_MS),
            ("grpc.keepalive_timeout_ms", KEEPALIVE_TIMEOUT_MS),
            ("grpc.http2.max_pings_without_data", HTTP2_MAX_PINGS_WITHOUT_DATA),
            ("grpc.keepalive_permit_without_calls", KEEPALIVE_PERMIT_WITHOUT_CALLS),
            ("grpc.lb_policy_name", "round_robin"),
        ]

    @staticmethod
    def _parse_endpoint_and_tls(host: str):
        parsed = urlparse(host)
        location = parsed.netloc.strip() or parsed.path
        return location, parsed.scheme == "https"

    def __init__(
        self,
        host: str,
        ServiceStub,
        max_message_length: int = DEFAULT_MAX_MESSAGE_LENGTH,
        request_compression_algo: str = "gzip",
        response_compression_algo: str = "gzip",
    ):
        self._server_address = host
        self._stub_cls = ServiceStub
        self._compression_in = request_compression_algo
        self._compression_out = response_compression_algo
        self._base_metadata = (("response-compression", response_compression_algo),)
        self._channel_opts = BasicAsyncClient._channel_options(max_message_length)
        self._endpoint, self._use_tls = BasicAsyncClient._parse_endpoint_and_tls(host)
        self._open_channels = {}

    async def _invoke(self, channel, method_name: str, request, metadata, timeout, stream: bool):
        stub = self._stub_cls(channel)
        stub_method = getattr(stub, method_name)
        if stream:
            async for response_message in stub_method(request, metadata=metadata, timeout=timeout):
                yield response_message
        else:
            yield await stub_method(request, metadata=metadata, timeout=timeout)

    async def send(
        self,
        request,
        Method: str,
        metadata: Tuple[Tuple[str, str], ...] = (),
        max_message_length: int = DEFAULT_MAX_MESSAGE_LENGTH,
        timeout=None,
        stream: bool = False,
    ):
        request_metadata = metadata + self._base_metadata
        channel_options = BasicAsyncClient._channel_options(max_message_length) if max_message_length != DEFAULT_MAX_MESSAGE_LENGTH else self._channel_opts
        if self._use_tls:
            ssl_credentials = grpc.ssl_channel_credentials()
            compression_algorithm = COMPRESSION.get(self._compression_in, grpc.Compression.Gzip)
        else:
            ssl_credentials = None
            compression_algorithm = COMPRESSION.get(self._compression_in, grpc.Compression.Gzip)

        if not stream:
            if self._use_tls:
                async with grpc.aio.secure_channel(self._endpoint, ssl_credentials, options=channel_options, compression=compression_algorithm) as channel:
                    response_generator = self._invoke(channel, Method, request, request_metadata, timeout, False)
                    return await response_generator.__anext__()
            async with grpc.aio.insecure_channel(self._endpoint, options=channel_options, compression=compression_algorithm) as channel:
                response_generator = self._invoke(channel, Method, request, request_metadata, timeout, False)
                return await response_generator.__anext__()

        if self._use_tls:
            channel = grpc.aio.secure_channel(self._endpoint, ssl_credentials, options=channel_options, compression=compression_algorithm)
        else:
            channel = grpc.aio.insecure_channel(self._endpoint, options=channel_options, compression=compression_algorithm)
        channel_identifier = id(channel)
        self._open_channels[channel_identifier] = channel

        async def stream_response_generator():
            try:
                async for response_message in self._invoke(channel, Method, request, request_metadata, timeout, True):
                    yield response_message
            finally:
                self._open_channels.pop(channel_identifier, None)

        return stream_response_generator()



import requests

import pytest

from redup_servicekit.monitoring import MonitorServer

def test_running():
    MonitorServer().run()
    monitoring = MonitorServer.get_instance()
    assert monitoring is not None

def test_prometheus():
    prometheus_lines = requests.get(
        'http://localhost:9999/metrics'
    ).text.split('\n')

    keys = set()
    for prometheus_line in prometheus_lines:
        prometheus_line = prometheus_line.strip()
        if not prometheus_line.startswith('#') and len(prometheus_line) > 0:
            keys.add(prometheus_line.split()[0].split('{')[0])

    for key in ['started_total', 'completed_total']:
        assert key in keys

@pytest.mark.asyncio
async def test_stats():
    monitoring = MonitorServer.get_instance()
    assert len(await monitoring.get_stats()) == 0



import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

import yaml


class ConfigSingleton:
    _data: Optional[Dict[str, Any]] = None

    @staticmethod
    def _deep_merge(source: Dict[str, Any], target: Dict[str, Any]) -> Dict[str, Any]:
        for config_key, config_value in source.items():
            if (
                isinstance(config_value, dict)
                and config_key in target
                and isinstance(target[config_key], dict)
            ):
                ConfigSingleton._deep_merge(config_value, target[config_key])
            else:
                target[config_key] = config_value
        return target

    @classmethod
    def load(cls, config_path: Union[str, Path]) -> Dict[str, Any]:
        config_path_resolved = Path(config_path)
        if not config_path_resolved.exists():
            raise ValueError(
                "The path to the configuration file does not exist.")
        if config_path_resolved.suffix != ".yaml":
            raise ValueError(
                "The path to the configuration file must be specified as a .yaml file."
            )
        with config_path_resolved.open() as config_file:
            cls._data = yaml.load(config_file, Loader=yaml.Loader)
        return cls._data

    @classmethod
    def get(cls) -> Dict[str, Any]:
        if cls._data is None:
            raise ValueError("Configuration not initialized.")
        return cls._data

    @classmethod
    def merge(cls, config: Union[str, Path, Dict[str, Any]]) -> None:
        if cls._data is None:
            raise ValueError("Configuration not initialized.")
        if isinstance(config, dict):
            cls._deep_merge(config, cls._data)
            return
        merge_path = Path(config)
        if not merge_path.exists():
            raise ValueError(
                "The path to the merge configuration file does not exist and is not a dictionary."
            )
        with merge_path.open() as merge_config_file:
            extra = yaml.load(merge_config_file, Loader=yaml.Loader)
        cls._deep_merge(extra, cls._data)

    @classmethod
    def inject_os_envs(cls) -> None:
        if cls._data is None:
            return
        changed = False
        for environment_variable_name in os.environ:
            path_segments = environment_variable_name.split("___")
            current_config_node = cls._data
            for path_segment in path_segments[:-1]:
                nested_key_with_spaces = path_segment.replace("__", " ")
                if nested_key_with_spaces not in current_config_node or not isinstance(
                    current_config_node.get(nested_key_with_spaces), dict
                ):
                    continue
                current_config_node = current_config_node[nested_key_with_spaces]
            leaf_config_key = path_segments[-1].replace("__", " ")
            if leaf_config_key not in current_config_node:
                continue
            environment_variable_value = os.environ[environment_variable_name]
            logging.info(
                "Changing value from config %s to environment value: %s.",
                repr(environment_variable_name),
                environment_variable_value,
            )
            current_config_node[leaf_config_key] = (
                environment_variable_value.split(",")
                if "," in environment_variable_value
                else environment_variable_value
            )
            changed = True
        if changed:
            logging.info("Current configuration: %s.", cls.get())
